from dataclasses import asdict, is_dataclass
from datetime import datetime
from json import JSONEncoder, dumps

import numpy as np


class CustomJSONEncoder(JSONEncoder):
    def default(self, o):
        if is_dataclass(o):
            return asdict(o)
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, datetime):
            return o.isoformat()
        if np.iscomplexobj(o):
            return str(o)
        return super().default(o)

    @classmethod
    def dumps(cls, o):
        return dumps(o, cls=cls)



from pathlib import Path

import numpy as np
import pytest
from aiosqlite import Connection, connect

from lisa.globalfit.framework.catalogs.sqlite import SqliteSourceCatalog
from lisa.globalfit.framework.model import MarkovChainSamples, ParametricModel
from lisa.globalfit.framework.msg.data import ModuleState


@pytest.fixture
def sample_module_states() -> dict[str, ModuleState]:
    def mock_module_state(model_name: str, params: np.ndarray) -> ModuleState:
        unused = ParametricModel(name="UNUSED", parameters=np.array([]))
        return ModuleState(
            chain=MarkovChainSamples(
                idx_iteration=0,
                idx_step_end=0,
                waveforms=ParametricModel(name=model_name, parameters=params),
            ),
            log_prior=unused,
            log_likelihood=unused,
            log_proposal=unused,
        )

    return {
        "A": mock_module_state("ModelA", np.array([[1, 2]])),
        "B": mock_module_state("ModelB", np.array([[3, 4]])),
    }


@pytest.fixture
async def sample_db():
    sample_data = [
        ("A", "2024-01-01 12:00:00+00:00", 1, 2),
        ("B", "2024-01-01 12:00:00+00:00", 3, 4),
        ("A", "2024-01-02 12:00:00+00:00", 5, 6),
        ("A", "2024-01-02 12:00:00+00:00", 7, 8),
    ]
    async with connect(":memory:") as db:
        await db.execute(
            """
            CREATE TABLE test_table (
                subject TEXT NOT NULL,
                timestamp TIMESTAMP NOT NULL,
                p0 INTEGER NOT NULL,
                p1 INTEGER NOT NULL
            )
            """
        )
        await db.executemany("INSERT INTO test_table VALUES (?, ?, ?, ?)", sample_data)
        await db.commit()
        yield db


@pytest.mark.asyncio
async def test_catalog_select_latest_model(sample_db: Connection, tmp_path: Path):
    result = await SqliteSourceCatalog(tmp_path).select_latest_model(
        sample_db, "test_table"
    )
    # Check that the two latest simulatneous updates from A are correctly concatenated.
    expected_result = {
        "A": ParametricModel(name="test_table", parameters=np.array([[5, 6], [7, 8]])),
        "B": ParametricModel(name="test_table", parameters=np.array([[3, 4]])),
    }
    assert result == expected_result


@pytest.mark.asyncio
async def test_catalog_roundtrip(
    sample_module_states: dict[str, ModuleState], tmp_path: Path
):
    # Make sure that sample data is not empty.
    assert sample_module_states

    obj = SqliteSourceCatalog(tmp_path / "catalog.db")
    for subject, state in sample_module_states.items():
        await obj.update(subject, state)
    latest = await obj.get_source_parameters()
    assert latest == {
        module: state.chain.waveforms for module, state in sample_module_states.items()
    }



import logging
from pathlib import Path

from aiohttp.web import (
    Application,
    GracefulExit,
    _run_app,
    normalize_path_middleware,
    static,
)

from lisa.globalfit.framework.monitoring import api, ui

logger = logging.getLogger(__name__)


class MonitoringServer:
    def __init__(self, api_url: str, s3_bucket_name: str) -> None:
        self.app = Application(middlewares=[normalize_path_middleware()])
        api.register(self.app, "/api", s3_bucket_name)
        ui.register(self.app, "/app", api_url)
        self.app.add_routes([static("/static", Path(__file__).parent / "static")])

    async def run(self, port: int) -> None:
        try:
            await _run_app(self.app, port=port)
        except (GracefulExit, KeyboardInterrupt):
            logger.info("stopping server")



import asyncio
import logging
from asyncio import Event, Future
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor

from lisa.globalfit.framework.bus import EventBus
from lisa.globalfit.framework.engine.rule import Response, Rule
from lisa.globalfit.framework.msg.subjects import Subject

logger = logging.getLogger(__name__)


class Engine:
    def __init__(self, bus: EventBus, rules: list[Rule]) -> None:
        self.bus = bus
        self.rules = rules
        self.handlers: dict[Subject, list[Rule]] = defaultdict(list)
        self.wake = Event()
        self.should_stop = Event()
        self.executor = ProcessPoolExecutor()
        self.running_tasks: set[Future] = set()

    async def setup(self) -> None:
        await self.bus.connect()
        await self.register_rules()

    async def register_rules(self) -> None:
        await asyncio.gather(*(self.register_rule(r) for r in self.rules))

    async def register_rule(self, rule: Rule) -> None:
        await self.bus.subscribe(rule.subject, self.handle_message)
        self.handlers[rule.subject].append(rule)
        logger.info(f"loaded rule {rule.id!r}")

    async def run(self) -> None:
        logger.info("starting rule engine")
        while not self.should_stop.is_set():
            if not self.running_tasks:
                logger.info("waiting for next event")
                # Relying on a wake event is necessary because we can't wait on an
                # empty list of coroutines.
                await self.wake.wait()

            self.wake.clear()
            logger.info(f"waiting for {len(self.running_tasks)} pending tasks")
            done, _ = await asyncio.wait(
                self.running_tasks, return_when=asyncio.FIRST_COMPLETED
            )
            self.running_tasks -= done
            await asyncio.gather(*(self.handle_result(result) for result in done))

        await self.shutdown()

    async def handle_result(self, task: Future) -> None:
        error = task.exception()
        if error is not None:
            logger.error(f"got exception for task {task}: {error}")
            return

        responses: list[Response] = task.result()
        await asyncio.gather(
            *(
                self.bus.publish(response.subject, response.msg.encode())
                for response in responses
            )
        )

    async def handle_message(self, subject: Subject, msg: bytes) -> None:
        handlers = self.handlers.get(subject)
        if handlers is None:
            logger.warning(f"no handlers for message on subject {subject!r}")
            return

        await asyncio.gather(
            *(self.run_handler(handler, subject, msg) for handler in handlers)
        )

    async def run_handler(self, rule: Rule, subject: Subject, msg: bytes) -> None:
        if not rule.is_active:
            logger.info(f"skipping evaluation of inactive rule {rule.id!r}")
            return

        logger.info(f"scheduling evaluation of rule {rule.id!r}")
        loop = asyncio.get_event_loop()
        task = loop.run_in_executor(self.executor, rule.evaluate, subject, msg)
        self.running_tasks.add(task)
        self.wake.set()

    async def publish_responses(self, responses: list[Response]) -> None:
        await asyncio.gather(
            *(self.bus.publish(r.subject, r.msg.encode()) for r in responses)
        )

    async def shutdown(self) -> None:
        logger.info("shutting down engine")
        await self.bus.close()
        self.executor.shutdown()



import json
import logging
import shutil
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from asyncio import get_event_loop, new_event_loop
from multiprocessing import Process
from pathlib import Path
from sys import exit
from typing import Optional

from lisa.globalfit.framework.args import (
    add_bus_args,
    add_common_args,
    add_engine_execution_args,
    add_execution_plot_args,
    add_module_execution_args,
    add_modules_execution_args,
    add_pipeline_args,
    add_preprocessing_args,
)
from lisa.globalfit.framework.buses.nats import NatsEventBus
from lisa.globalfit.framework.catalogs.memory import InMemorySourceCatalog
from lisa.globalfit.framework.catalogs.sqlite import SqliteSourceCatalog
from lisa.globalfit.framework.exceptions import FrameworkError
from lisa.globalfit.framework.mock_engine import Engine, ModulePool
from lisa.globalfit.framework.modules.registry import ModuleRegistry
from lisa.globalfit.framework.modules.stub.preprocessor import StubPreprocessor
from lisa.globalfit.framework.msg.control import (
    ConfigureModule,
    IterateModule,
    ModuleGroupConfiguration,
)
from lisa.globalfit.framework.msg.subjects import channel_configure, create_channel_name
from lisa.globalfit.framework.plot import ModuleExecutionDrawer

logger = logging.getLogger(__name__)


def default_workdir(dir: Optional[Path]) -> Path:
    return dir or Path.cwd()


def get_args() -> Namespace:
    parser = ArgumentParser(
        "LISA GlobalFit Framework CLI",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    add_common_args(parser)
    add_bus_args(parser)
    add_pipeline_args(parser)

    subparsers = parser.add_subparsers(dest="command", required=True)
    add_preprocessing_args(subparsers)
    add_modules_execution_args(subparsers)
    add_module_execution_args(subparsers)
    add_engine_execution_args(subparsers)
    add_execution_plot_args(subparsers)

    return parser.parse_args()


async def cmd_preprocess(args: Namespace):
    preprocessor = StubPreprocessor(args.group_label)
    workdir = default_workdir(args.workdir)
    datafile = workdir / args.input_data
    checkpoint = (
        ModulePool.from_json(workdir / args.input_checkpoint)
        if args.input_checkpoint
        else None
    )
    bus = NatsEventBus(args.servers)
    await bus.connect()
    # TODO Refactor subject creation with engine code.
    channel = create_channel_name(args.pipeline_run_id, channel_configure("engine"))
    logger.info(f"running {preprocessor.group_label} preprocessor")
    config = preprocessor.preprocess(datafile, checkpoint)
    preprocessor.write_group_configuration(config, args.output_group_configuration_path)

    # TODO If checkpoint is given, subtract signal from other groups to reduce the
    # amount of pre-processing to be done by children modules.
    # For now, the input data file is copied as-is to the preprocessed signal
    # location.
    logger.warning("TODO: group-level subtraction of signals")
    shutil.copy(datafile, args.output_preprocessed_signal)

    await bus.publish(
        channel,
        ModuleGroupConfiguration(
            group=preprocessor.group_label,
            labels=tuple(worker.module_label for worker in config.modules),
        ).encode(),
    )


async def cmd_plot(args: Namespace):
    drawer = ModuleExecutionDrawer()
    drawer.plot_corner(
        chains=args.input_chain,
        dataset=args.chain_location,
        output=Path(args.input_chain).with_suffix(".svg"),
    )


async def cmd_start_module(args: Namespace):
    workdir = default_workdir(args.workdir)
    # TODO Split all stub-related code to dedicated package.
    group = ModulePool.from_group_configuration(
        args.input_group_configuration_path, args.pipeline_run_id
    )
    channel_name = create_channel_name(
        args.pipeline_run_id,
        args.group_label,
        args.module_label,
    )
    module = ModuleRegistry.create(
        name="stub",
        datafile=workdir / args.input_data,
        output_checkpoint=workdir / args.output_checkpoint,
        channel_name=channel_name,
        bus=NatsEventBus(args.servers),
    )
    iteration_config = IterateModule(
        idx_iteration=args.current_iteration,
        detections=InMemorySourceCatalog(
            detection_map={
                subject: entry.state.log_likelihood
                for subject, entry in group.workers.items()
                if entry.state.log_likelihood.parameters
            }
        ),
        step_count=args.n_steps,
    )
    await module.setup()
    await module.run(
        ConfigureModule(group.workers[channel_name].state), iteration_config
    )


async def cmd_start_all_modules(args: Namespace):
    pool = ModulePool.from_group_configuration(args.input_group_configuration_path)
    pids = [
        Process(
            target=main,
            args=(
                # Forward arguments to module subprocess, overriding the command and
                # setting the corresponding module label.
                Namespace(
                    **{
                        **vars(args),
                        "no_global_loop": True,
                        "command": "start-module",
                        "module_label": module.label,
                        "group_label": module.group,
                        "output_checkpoint": f"checkpoint_{module.label}.h5",
                    }
                ),
            ),
        )
        for module in pool.workers.values()
    ]

    logger.info(f"starting {len(pids)} module subprocesses")
    for pid in pids:
        pid.start()

    for pid in pids:
        pid.join()
        if pid.exitcode != 0:
            raise FrameworkError(f"subprocess {pid} exited with error")


async def cmd_start_engine(args: Namespace):
    with open(args.expected_groups) as f:  #  noqa: ASYNC230
        content = json.load(f)

    engine = Engine(
        expected_groups=[group["label"] for group in content],
        bus=NatsEventBus(args.servers),
        catalog=SqliteSourceCatalog(args.catalog_db),
        output=args.output_checkpoint,
        pipeline_run_id=args.pipeline_run_id,
        max_iterations=args.max_iterations,
    )
    await engine.setup()
    result = await engine.run(args.current_iteration)
    await engine.write_iteration_decision(result, args.output_iteration_decision)


def main(args: Optional[Namespace] = None):
    args = args or get_args()
    logging.basicConfig(
        level=args.log_level,
        format="%(asctime)s.%(msecs)03d [%(levelname)s] %(name)s %(message)s",
        datefmt=r"%Y-%m-%dT%H:%M:%S",
    )

    cmd_registry = {
        "plot": cmd_plot,
        "preprocess": cmd_preprocess,
        "start-module": cmd_start_module,
        "start-modules": cmd_start_all_modules,
        "start-engine": cmd_start_engine,
    }
    fn = cmd_registry.get(args.command)

    if fn is None:
        logger.error(f"unknown command {args.command}")
        exit(1)

    # When spawning subprocesses, the global event loop is used by default. This raises
    # runtime errors when trying to run tasks because the event loop is already
    # running. To prevent this behaviour, we create a new event loop in this specific
    # case.
    # Note that this is only required for development purposes which rely on the
    # `start-modules` command, but in production containers will be started
    # independently.
    # See https://github.com/python/asyncio/issues/421
    if vars(args).get("no_global_loop", False):
        loop = new_event_loop()
    else:
        loop = get_event_loop()

    task = loop.create_task(fn(args))
    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt:
        # TODO Add two-stages cleanup mechanism, waiting for in-flight tasks to
        # complete, falling back to direct exit on second KeyboardInterrupt.
        logger.info("exiting immediately")
        exit(1)
    except FrameworkError as err:
        logger.error(err)
        exit(1)


if __name__ == "__main__":
    main()



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from aiobotocore.session import AioSession, get_session
from aiohttp.web import (
    AppKey,
    Application,
    Request,
    Response,
    RouteTableDef,
    json_response,
)
from types_aiobotocore_s3.client import S3Client

from lisa.globalfit.framework.io import CustomJSONEncoder
from lisa.globalfit.framework.monitoring import service
from lisa.globalfit.framework.monitoring.types import MonitoringStorage

# Use typed application-level variables.
# See https://docs.aiohttp.org/en/stable/web_advanced.html#application-s-config
s3_session = AppKey("s3_session", AioSession)
s3_bucket = AppKey("s3_bucket", str)


routes = RouteTableDef()


@asynccontextmanager
async def get_s3_client(request: Request) -> AsyncIterator[S3Client]:
    # Use AWS_ENDPOINT_URL, AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY environment
    # variables for client authentication.
    async with request.app[s3_session].create_client(
        "s3",
    ) as client:
        yield client


@asynccontextmanager
async def get_monitoring_storage(request: Request) -> AsyncIterator[MonitoringStorage]:
    async with get_s3_client(request) as s3:
        yield MonitoringStorage(s3, request.app[s3_bucket])


@routes.get("/runs")
async def get_runs(request: Request) -> Response:
    async with get_monitoring_storage(request) as storage:
        response = await service.get_runs(storage)
    return json_response(response, dumps=CustomJSONEncoder.dumps)


@routes.get("/runs/{run}/groups")
async def get_run_groups(request: Request) -> Response:
    async with get_monitoring_storage(request) as storage:
        response = await service.get_run_groups(storage, request.match_info["run"])
    return json_response(response, dumps=CustomJSONEncoder.dumps)


@routes.get("/runs/{run}/groups/{group}/modules")
async def get_run_group_modules(request: Request) -> Response:
    async with get_monitoring_storage(request) as storage:
        response = await service.get_run_group_modules(
            storage, request.match_info["run"], request.match_info["group"]
        )
    return json_response(response, dumps=CustomJSONEncoder.dumps)


@routes.get("/runs/{run}/groups/{group}/modules/{module}")
async def get_run_group_module(request: Request) -> Response:
    async with get_monitoring_storage(request) as storage:
        response = await service.get_run_group_module(
            storage,
            request.match_info["run"],
            request.match_info["group"],
            request.match_info["module"],
        )
    return json_response(response, dumps=CustomJSONEncoder.dumps)


def register(app: Application, prefix: str, s3_bucket_name: str):
    subapp = Application()
    subapp.add_routes(routes)

    # Configure S3 client.
    subapp[s3_session] = get_session()
    subapp[s3_bucket] = s3_bucket_name
    app.add_subapp(prefix, subapp)



import asyncio
import json
import logging
from collections.abc import Awaitable
from dataclasses import dataclass
from pathlib import Path
from typing import Callable

from lisa.globalfit.framework.bus import EventBus
from lisa.globalfit.framework.catalog import SourceCatalog
from lisa.globalfit.framework.exceptions import DeserializationError, FrameworkError
from lisa.globalfit.framework.io import CustomJSONEncoder
from lisa.globalfit.framework.modules.module import ModuleExecutionState
from lisa.globalfit.framework.modules.pool import ModulePool, ModulePoolEntry
from lisa.globalfit.framework.msg.base import MessageBase
from lisa.globalfit.framework.msg.control import (
    ExecutionStateUpdate,
    ModuleGroupConfiguration,
    TerminateModule,
)
from lisa.globalfit.framework.msg.data import ModuleState
from lisa.globalfit.framework.msg.subjects import (
    Subject,
    channel_any_group_any_module,
    channel_configure,
    channel_data_state,
    channel_status,
    channel_strip,
    channel_terminate,
    create_channel_name,
)

logger = logging.getLogger(__name__)


@dataclass
class IterationDecision:
    continue_iterating: bool


class Engine:
    DEFAULT_EVENTLOOP_SLEEP_S = 1

    def __init__(
        self,
        expected_groups: list[str],
        bus: EventBus,
        catalog: SourceCatalog,
        output: Path,
        pipeline_run_id: str,
        max_iterations: int,
    ) -> None:
        self.pipeline_run_id = pipeline_run_id
        self.expected_groups = set(expected_groups)
        self.actual_groups: set[str] = set()
        self.bus = bus
        self.catalog = catalog
        self.output = output
        self.max_iterations = max_iterations
        self.pool = ModulePool(workers={})

    async def setup(self) -> None:
        await self.bus.connect()
        # Subscribe to preprocessors updates.
        await self.set_safe_callback(
            subject=create_channel_name(
                self.pipeline_run_id, channel_configure("engine")
            ),
            msg_type=ModuleGroupConfiguration,
            fn=self.handle_group_configuration,
        )
        # Subscribe to all module updates.
        await self.set_safe_callback(
            subject=channel_any_group_any_module(self.pipeline_run_id, channel_status),
            msg_type=ExecutionStateUpdate,
            fn=self.pool.update_module_state,
        )
        await self.set_safe_callback(
            subject=channel_any_group_any_module(
                self.pipeline_run_id, channel_data_state
            ),
            msg_type=ModuleState,
            fn=self.handle_module_data,
        )

    async def shutdown(self) -> None:
        await self.bus.close()

    async def run(
        self,
        current_iteration: int,
    ) -> IterationDecision:
        if not self.expected_groups:
            raise FrameworkError("engine needs to expect at least one group")

        logger.info("starting engine")
        try:
            logger.info(
                f"running iteration {current_iteration}/{self.max_iterations-1}"
            )
            await self.wait_for_groups_configurations()
            await self.wait_for_modules_iteration(current_iteration)
        except FrameworkError:
            logger.error("shutting down engine")
            raise
        finally:
            await self.catalog.write(self.output.parent / "catalog.json")
            await self.shutdown_pool()
            await self.shutdown()
            return IterationDecision(
                continue_iterating=current_iteration < self.max_iterations - 1
            )

    async def write_iteration_decision(
        self, decision: IterationDecision, filename: Path
    ) -> None:
        with open(filename, "w") as f:  #  noqa: ASYNC230
            json.dump(decision, f, cls=CustomJSONEncoder)
        logger.info(f"wrote iteration decision to {filename}")

    async def wait_for_groups_configurations(
        self, check_interval: int = DEFAULT_EVENTLOOP_SLEEP_S
    ):
        while self.expected_groups != self.actual_groups:
            diff = self.expected_groups - self.actual_groups
            logger.info(f"waiting for {len(diff)} configurations: {diff}")
            await asyncio.sleep(check_interval)

    async def handle_group_configuration(
        self, subject: Subject, msg: ModuleGroupConfiguration
    ) -> None:
        # Perform base sanity checks.
        if msg.group not in self.expected_groups:
            logger.warning(f"got unexpected configuration for group {msg.group}")
            return
        if msg.group in self.actual_groups:
            logger.warning(f"got duplicate configuration for group {msg.group}")
            return

        new_entries = {
            create_channel_name(
                self.pipeline_run_id, msg.group, label
            ): ModulePoolEntry(
                group=msg.group,
                label=label,
                execution_state=ModuleExecutionState.CREATED,
                # TODO Should state be optional ?
                state=None,  # type: ignore
            )
            for label in msg.labels
        }
        self.pool.workers |= new_entries
        self.actual_groups.add(msg.group)
        logger.info(
            f"got configuration of {len(new_entries)} modules for group {msg.group}"
        )

    async def wait_for_modules_iteration(
        self,
        iteration_number: int,
        check_interval: int = DEFAULT_EVENTLOOP_SLEEP_S,
    ):
        while remaining := self.pool.select_modules(
            lambda module: not (
                module.execution_state == ModuleExecutionState.IDLE
                and module.current_iteration == iteration_number
                and module.state is not None
            )
        ):
            progress = f"{len(self.pool) - len(remaining)}/{len(self.pool)}"
            logger.info(f"{progress} modules iterated")
            logger.debug(f"current module pool state: {self.pool.workers}")
            await asyncio.sleep(check_interval)

        self.pool.write_checkpoint(self.output)
        logger.info("all modules finished iteration")

    async def shutdown_pool(self) -> None:
        await self.send_group(self.send_terminate, list(self.pool.workers))

    async def send_group(
        self, fn: Callable[[Subject], Awaitable[None]], subjects: list[Subject]
    ) -> None:
        await asyncio.gather(*(fn(s) for s in subjects))

    async def send_terminate(self, subject: Subject) -> None:
        logger.debug(f"sending terminate message to {subject}")
        msg = TerminateModule()
        await self.bus.publish(channel_terminate(subject), msg.encode())

    async def set_safe_callback(
        self, subject: Subject, msg_type: type[MessageBase], fn: Callable
    ) -> None:
        async def handler(subject: Subject, msg: bytes):
            try:
                decoded_msg = msg_type.decode(msg)
            except DeserializationError as err:
                # Discard unknown messages.
                logger.error(f"could not read message from {subject}: {err}")
                return

            module = channel_strip(subject)
            await fn(module, decoded_msg)

        await self.bus.subscribe(subject, handler)

    async def handle_module_data(self, subject: Subject, state: ModuleState):
        await self.catalog.update(subject, state)
        await self.pool.update_global_state(subject, state)



import asyncio
import logging
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from asyncio.exceptions import CancelledError
from pathlib import Path
from signal import SIGINT, SIGTERM

from lisa.globalfit.framework.args import add_bus_args, add_common_args
from lisa.globalfit.framework.buses.nats import NatsEventBus
from lisa.globalfit.framework.engine.engine import Engine
from lisa.globalfit.framework.engine.rule import Rule

logger = logging.getLogger(__name__)


def get_args() -> Namespace:
    parser = ArgumentParser(
        "LISA GlobalFit Framework Engine CLI",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    add_common_args(parser)
    add_bus_args(parser)

    rule_selection_group = parser.add_mutually_exclusive_group(required=True)
    rule_selection_group.add_argument(
        "--rules-directory",
        "-d",
        help="Directory containing rules",
        type=Path,
    )
    rule_selection_group.add_argument(
        "--rule",
        "-f",
        help="File containing individual rule, can be passed multiple times",
        type=Path,
        nargs="+",
    )

    return parser.parse_args()


async def run_engine(args: Namespace):
    if args.rules_directory:
        files = Path(args.rules_directory).rglob("*.py")
    else:
        files = args.rule

    rules = [Rule.from_file(f) for f in files]
    engine = Engine(
        bus=NatsEventBus(args.servers),
        rules=rules,
    )
    try:
        await engine.setup()
        await engine.run()
    except CancelledError:
        await engine.shutdown()


def main():
    args = get_args()
    logging.basicConfig(
        level=args.log_level,
        format="%(asctime)s.%(msecs)03d [%(levelname)s] %(name)s %(message)s",
        datefmt=r"%Y-%m-%dT%H:%M:%S",
    )

    loop = asyncio.get_event_loop()
    if logging.root.level == logging.DEBUG:
        loop.set_debug(True)
    task = asyncio.ensure_future(run_engine(args))

    for signal in [SIGINT, SIGTERM]:
        loop.add_signal_handler(signal, task.cancel)

    loop.run_until_complete(task)


if __name__ == "__main__":
    main()



import logging
from dataclasses import dataclass
from typing import ClassVar

from lisa.globalfit.framework.modules.module import ModuleBase

logger = logging.getLogger(__name__)


@dataclass
class ModuleRegistry:
    items: ClassVar[dict[str, type[ModuleBase]]] = {}

    @classmethod
    def choices(cls) -> list[str]:
        return list(cls.items)

    @classmethod
    def register(cls, name: str):
        def decorator(module_class: type[ModuleBase]):
            cls.items[name] = module_class
            return module_class

        return decorator

    @classmethod
    def create(cls, name: str, *args, **kwargs) -> ModuleBase:
        try:
            kls = cls.items[name]
            return kls(*args, **kwargs)
        except KeyError:
            logger.error(f"unknown module identifier: {name}")
            raise



from typing import Callable

Subject = str

SUBJECT_SCIENCE = Subject("sci")
SUBJECT_INFRA = Subject("infra")

SUBJECT_GLOBALFIT = Subject(f"{SUBJECT_SCIENCE}.gf")

PERSISTENT_SUBJECTS = [SUBJECT_SCIENCE, SUBJECT_INFRA]

ANY_SUBJECT = "*"


def create_channel_name(*args: str) -> Subject:
    return ".".join(
        [
            SUBJECT_SCIENCE,
            "gf",
            *args,
        ]
    )


def channel_any(channel_fn: Callable) -> Subject:
    return channel_fn(
        ".".join([SUBJECT_GLOBALFIT, ANY_SUBJECT, ANY_SUBJECT, ANY_SUBJECT])
    )


def channel_any_group_any_module(run_id: Subject, channel_fn: Callable) -> Subject:
    return channel_fn(".".join([SUBJECT_GLOBALFIT, run_id, ANY_SUBJECT, ANY_SUBJECT]))


def channel_get_pipeline(channel: Subject) -> str:
    return channel.split(".")[2]


def channel_get_group(channel: Subject) -> str:
    return channel.split(".")[3]


def channel_get_module(channel: Subject) -> str:
    return channel.split(".")[4]


def channel_strip(sub: Subject) -> Subject:
    return ".".join(sub.split(".")[:-2])


def channel_configure(sub: Subject) -> Subject:
    return f"{sub}.control.configure"


def channel_iterate(sub: Subject) -> Subject:
    return f"{sub}.control.iterate"


def channel_terminate(sub: Subject) -> Subject:
    return f"{sub}.control.terminate"


def channel_status(sub: Subject) -> Subject:
    return f"{sub}.control.status"


def channel_data_state(sub: Subject) -> Subject:
    return f"{sub}.data.state"



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum, auto

import numpy as np
import numpy.typing as npt
from typing_extensions import Self


@dataclass
class ParametricModel:
    """Well-known parametric model.

    :param name: Identifier of the model.
    :param parameters: Parameters of the model.
    """

    name: str
    parameters: npt.NDArray

    @classmethod
    def from_dict(cls, obj: dict) -> Self:
        return cls(name=obj["name"], parameters=np.array(obj["parameters"]))

    def __eq__(self, other) -> bool:
        """Check whether two models are equivalent.

        Use with caution!

        This is more of a helper for unit tests than a real feature, since it performs
        equality checks on floating point arrays.

        """
        if not isinstance(other, ParametricModel):
            return False
        if self.name != other.name:
            return False
        return np.array_equal(self.parameters, other.parameters)


@dataclass
class MarkovChainSamples:
    """Sequence of Markov chain samples.

    :param idx_iteration: Current Global Fit iteration.
    :param: idx_step_end: Step of the last element in waveforms parameters.
    :param waveforms: Model of the waveform. Its name represents the function identifier
        of the associated waveform generator. Its parameters represent the samples and
        have a shape of n_steps * n_walkers * n_dim. Its size is estimated to be of the
        following order of magnitude.

             1000 (steps) * 20 (parameters) * 5 (walkers) * 4 (f32) ~ 400kB.
    """

    idx_iteration: int
    idx_step_end: int
    waveforms: ParametricModel

    @classmethod
    def from_dict(cls, obj: dict) -> Self:
        return cls(
            idx_iteration=obj["idx_iteration"],
            idx_step_end=obj["idx_step_end"],
            waveforms=ParametricModel.from_dict(obj["waveforms"]),
        )


class ModuleExecutionState(IntEnum):
    # TODO Use state machine if we need to handle complex transitions.
    CREATED = auto()
    CONFIGURING = auto()
    IDLE = auto()
    RUNNING = auto()
    ERROR = auto()
    DONE = auto()



from __future__ import annotations

import logging
from abc import abstractmethod
from os import getenv
from pathlib import Path
from typing import Callable

from aiobotocore.session import get_session
from botocore.exceptions import ClientError

from lisa.globalfit.framework import __version__
from lisa.globalfit.framework.bus import EventBus
from lisa.globalfit.framework.exceptions import DeserializationError, FrameworkError
from lisa.globalfit.framework.model import ModuleExecutionState
from lisa.globalfit.framework.msg.base import MessageBase
from lisa.globalfit.framework.msg.control import (
    ConfigureModule,
    ExecutionStateUpdate,
    IterateModule,
    TerminateModule,
)
from lisa.globalfit.framework.msg.data import ModuleState
from lisa.globalfit.framework.msg.subjects import (
    Subject,
    channel_data_state,
    channel_get_group,
    channel_get_module,
    channel_get_pipeline,
    channel_status,
    channel_terminate,
)

logger = logging.getLogger(__name__)


class ModuleBase:
    """Abstract Global Fit module."""

    DEFAULT_POLL_INTERVAL_S = 1
    DEFAULT_S3_BUCKET_VAR = "S3_BUCKET"
    DEFAULT_S3_PREFIX = "lisa-checkpoint"

    def __init__(
        self, datafile: Path, output_checkpoint: Path, channel_name: str, bus: EventBus
    ) -> None:
        """Create base module.

        :param datafile: Input data file containing signal.
        :param output_checkpoint: Path to a save state, from which execution can be
            resumed after nominal termination or a crash.
        :param channel_name: Topic prefix to use when talking on the bus.
        :param bus: Event bus interface.
        """
        self.datafile = datafile
        self.output_checkpoint = output_checkpoint
        self.channel_name = channel_name
        self.bus = bus
        self.execution_state = ModuleExecutionState.CREATED
        self.current_iteration = -1
        self.current_step = -1

        # When we will have dynamic reallocation of modules (#5), the container
        # orchestrator (Argo) will be able to upload checkpoints to S3 automatically
        # at each iteration.
        # However, for now modules live for the entire duration of the global fit run,
        # so they need to manually push their checkpoints to S3.
        # This is done using environment variables because it's a temporary workaround
        # which is not meant to appear in the public API.
        self._s3_bucket = getenv(self.DEFAULT_S3_BUCKET_VAR)
        if self._s3_bucket is None:
            logger.info(
                f"env variable {self.DEFAULT_S3_BUCKET_VAR} is not set, "
                "checkpoints will not be uploaded to S3"
            )

    async def setup(self) -> None:
        await self.bus.connect()
        await self.set_safe_callback(
            channel_terminate(self.channel_name),
            TerminateModule,
            self._terminate,
        )

    async def set_safe_callback(
        self, subject: Subject, msg_type: type[MessageBase], fn: Callable
    ) -> None:
        async def handler(subject: Subject, msg: bytes) -> None:
            # Make sure message can be decoded.
            try:
                decoded_msg = msg_type.decode(msg)
            except DeserializationError as err:
                # Discard unknown messages.
                logger.error(f"could not read message on {subject}: {err}")
                return

            logger.debug(f"{self.channel_name} received {type(decoded_msg).__name__}")

            # Run callback using decoded message.
            await fn(decoded_msg)

        await self.bus.subscribe(subject, handler)

    async def publish_results(self, results: ModuleState) -> None:
        await self.bus.publish(channel_data_state(self.channel_name), results.encode())

    async def update_status(self, status: ModuleExecutionState) -> None:
        self.execution_state = status
        await self.bus.publish(
            channel_status(self.channel_name),
            ExecutionStateUpdate(
                current_iteration=self.current_iteration, new_state=status
            ).encode(),
        )

    async def run(
        self, initial_state: ConfigureModule, iteration_config: IterateModule
    ):
        await self._configure(initial_state)
        await self._iterate(iteration_config)

    async def _configure(self, msg: ConfigureModule) -> None:
        if self.execution_state != ModuleExecutionState.CREATED:
            logger.warning(
                f"{self.channel_name} got {msg} but status is {self.execution_state}"
            )

        await self.update_status(ModuleExecutionState.CONFIGURING)
        await self.configure(msg)
        await self.update_status(ModuleExecutionState.IDLE)

    async def _iterate(self, msg: IterateModule) -> None:
        if self.execution_state != ModuleExecutionState.IDLE:
            logger.warning(
                f"{self.channel_name} got {msg} but status is {self.execution_state}"
            )

        self.current_iteration = msg.idx_iteration
        await self.update_status(ModuleExecutionState.RUNNING)
        try:
            state = await self.iterate(msg)
            # FIXME We need to get the current step count as an input argument, because
            # we don't know how many have been done in the past.
            self.current_step = msg.step_count
            state.checkpoint_key = self._derive_checkpoint_key()
            await self.publish_results(state)
            await self._publish_checkpoint(state.checkpoint_key)
            await self.update_status(ModuleExecutionState.IDLE)
        except FrameworkError as err:
            logger.error(f"{self.channel_name}: {err}")
            await self.update_status(ModuleExecutionState.ERROR)

    async def _terminate(self, msg: TerminateModule) -> None:
        logger.debug(f"stopping {self.channel_name}")
        self.execution_state = ModuleExecutionState.DONE

    def _derive_checkpoint_key(self, prefix=DEFAULT_S3_PREFIX) -> str:
        return "-".join((prefix, self.channel_name))

    async def _publish_checkpoint(self, object_name: str) -> None:
        if self._s3_bucket is not None:
            metadata = {
                "gf-version": __version__,
                "gf-pipeline-run": channel_get_pipeline(self.channel_name),
                "gf-group": channel_get_group(self.channel_name),
                "gf-module": channel_get_module(self.channel_name),
                "gf-iteration": str(self.current_iteration),
            }
            async with get_session().create_client("s3") as s3:
                try:
                    # TODO Use aiofiles to get rid of ASYNC230.
                    with open(self.output_checkpoint, "rb") as f:  #  noqa: ASYNC230
                        res = await s3.put_object(
                            Bucket=self._s3_bucket,
                            Key=object_name,
                            Body=f,
                            Metadata=metadata,
                        )
                    logger.debug(f"uploaded checkpoint: {res}")
                # FIXME This might not be the correct exception type for aiobotocore.
                except ClientError as err:
                    logger.warning(f"could not upload checkpoint: {err}")
        else:
            logger.debug("skipping checkpoint upload")

    @abstractmethod
    async def configure(self, msg: ConfigureModule) -> None:
        pass

    @abstractmethod
    async def iterate(self, msg: IterateModule) -> ModuleState:
        pass



import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import numpy as np

from lisa.globalfit.framework.model import MarkovChainSamples, ParametricModel
from lisa.globalfit.framework.modules.pool import ModulePool
from lisa.globalfit.framework.msg.data import ModuleState
from lisa.globalfit.framework.preprocessing import (
    ModuleConfiguration,
    PreprocessorOutput,
)
from lisa.globalfit.framework.signal import NoisyGaussianMixture

logger = logging.getLogger(__name__)

DUMMY_CONFIG_FILENAME = "dummy_config.json"


@dataclass
class StubPreprocessor:
    group_label: str

    def preprocess(
        self, dataset: Path, checkpoint: Optional[ModulePool] = None
    ) -> PreprocessorOutput:
        logger.debug(f"pre-processing dataset {dataset}")

        if checkpoint is not None:
            return self.preprocess_from_checkpoint(checkpoint)

        # Read input data.
        data = NoisyGaussianMixture.read_dataset(dataset)

        # Create module configurations, based on true values found in the catalog.
        return PreprocessorOutput(
            modules=[
                ModuleConfiguration(
                    group_label=self.group_label,
                    module_label=f"{data.means[i]:.0f}",
                    initial_state=ModuleState(
                        chain=MarkovChainSamples(
                            idx_iteration=0,
                            idx_step_end=0,
                            waveforms=ParametricModel(
                                name="norm",
                                # Cheat using the actual "catalog" value as initial
                                # state.
                                parameters=np.array(
                                    [[data.means[i], data.variances[i]]]
                                ),
                            ),
                        ),
                        # For now, we cheat by using the actual "catalog" values as
                        # prior.
                        log_prior=ParametricModel(
                            name="multivariate_normal",
                            # Only specify prior mean, use default covariance matrix.
                            parameters=np.array([data.means[i], data.variances[i]]),
                        ),
                        log_likelihood=ParametricModel("gaussian1d", np.array([])),
                        # TODO Use actual proposal distribution.
                        log_proposal=ParametricModel(
                            name="unused", parameters=np.array([])
                        ),
                    ),
                    static_configuration_reference=DUMMY_CONFIG_FILENAME,
                )
                for i in range(data.n_distributions())
            ]
        )

    def preprocess_from_checkpoint(self, checkpoint: ModulePool) -> PreprocessorOutput:
        group_workers = [
            entry
            for entry in checkpoint.workers.values()
            if entry.group == self.group_label
        ]
        logger.debug(
            f"got checkpoint with {len(group_workers)} {self.group_label} entries"
        )
        logger.info(group_workers)
        return PreprocessorOutput(
            modules=[
                # FIXME Propagate static configuration reference.
                ModuleConfiguration(
                    entry.group, entry.label, entry.state, DUMMY_CONFIG_FILENAME
                )
                for entry in group_workers
            ]
        )

    def write_group_configuration(
        self, config: PreprocessorOutput, output: Path
    ) -> None:
        config.write_json(output)



from typing import Callable

from lisa.globalfit.framework.msg.subjects import Subject


class EventBus:
    def __init__(self) -> None:
        pass

    async def connect(self) -> None:
        pass

    async def publish(self, subject: Subject, data: bytes) -> None:
        pass

    async def subscribe(self, subject: Subject, callback: Callable) -> None:
        pass

    async def unsubscribe(self, subject: Subject) -> None:
        pass

    async def close(self) -> None:
        pass



import logging

import numpy as np

from lisa.globalfit.framework.signal import NoisyGaussianMixture
from lisa.globalfit.framework.waveform import Gaussian1D

logger = logging.getLogger(__name__)


class Simulator:
    def __init__(self) -> None:
        self.rng = np.random.default_rng(seed=42)

    def simulate_gaussians(
        self,
        n_distributions: int,
        n_samples: int,
        snr: int,
    ) -> NoisyGaussianMixture:
        logger.info(f"simulating {n_distributions} gaussians with snr {snr}")

        # Generate time points.
        time_points = np.arange(n_samples)

        # Initialize the sum of Gaussian-shaped time series.
        sum_of_gaussians = np.zeros(n_samples)

        # Draw random mean values between start and end of signal.
        means = self.rng.uniform(0, n_samples, n_distributions)

        # Draw random standard deviations between normal and somewhat flattened.
        variances = self.rng.uniform(1, 10, n_distributions)

        # Generate each Gaussian-shaped time series and add to the sum.
        waveform = Gaussian1D()
        for i in range(n_distributions):
            # Generate values based on a Gaussian curve, add them to the sum.
            sum_of_gaussians += waveform.evaluate(time_points, means[i], variances[i])

        # Add noise to the signal.
        noise_factor = sum_of_gaussians.max() / snr
        noise = noise_factor * self.rng.standard_normal(n_samples)
        sum_of_gaussians += noise

        return NoisyGaussianMixture(
            means=means,
            variances=variances,
            noise=noise,
            signal=np.stack((time_points, sum_of_gaussians)),
        )



import json
import logging
import sqlite3
from asyncio import gather
from collections import ChainMap
from datetime import datetime, timezone
from itertools import groupby
from pathlib import Path
from textwrap import dedent

import numpy as np
from aiosqlite import Connection, connect

from lisa.globalfit.framework.catalog import SourceCatalog
from lisa.globalfit.framework.io import CustomJSONEncoder
from lisa.globalfit.framework.model import ParametricModel
from lisa.globalfit.framework.msg.data import ModuleState
from lisa.globalfit.framework.msg.subjects import Subject

sqlite3.register_adapter(np.float64, float)
sqlite3.register_adapter(np.float32, float)
sqlite3.register_adapter(np.int64, int)
sqlite3.register_adapter(np.int32, int)

logger = logging.getLogger(__name__)


class SqliteSourceCatalog(SourceCatalog):
    def __init__(self, db_file: Path) -> None:
        super().__init__()
        self.db_file = db_file

    async def get_source_parameters(self) -> dict[Subject, ParametricModel]:
        async with connect(self.db_file) as db:
            tables = await self.get_model_tables(db)
            source_parameters = await gather(
                *(self.select_latest_model(db, table) for table in tables)
            )
            return dict(ChainMap(*source_parameters))

    async def update(self, subject: Subject, state: ModuleState) -> None:
        # TODO Should the detection catalog be updated if the best sample from the
        # chain in the current iteration has a lower posterior probability than the
        # best sample from all previous iterations ?
        # The posterior should probably be passed along the waveform parameters, so
        # that its evolution can be monitored during execution.
        async with connect(self.db_file) as db:
            await db.set_trace_callback(logger.debug)
            # TODO Ideally, we should only do this once, but we need to figure out
            # corner cases.
            # Maybe we should try/except the insert directly, and only create the table
            # on error.
            await self.ensure_table(db, state.chain.waveforms)
            await self.insert_values(db, subject, state.chain.waveforms)

    async def write(self, filename: Path) -> None:
        params = await self.get_source_parameters()
        with open(filename, "w") as f:  #  noqa: ASYNC230
            json.dump(params, f, cls=CustomJSONEncoder)
        logger.info(f"wrote {len(params)} source parameters to {filename}")

    def safe_table_name(self, model_name: str) -> str:
        if not model_name.isidentifier():
            raise ValueError("")
        return model_name

    async def ensure_table(self, db: Connection, model: ParametricModel) -> None:
        # We cannot use prepared statements to parametrize table name, so we use
        # string interpolation on a safe version of the model name to prevent SQL
        # injection.
        table = self.safe_table_name(model.name)
        n_params = model.parameters.shape[-1]
        dtype = model.parameters.dtype
        params = ", ".join(f"p{i} {dtype} NOT NULL" for i in range(n_params))
        stmt = dedent(
            f"""
            CREATE TABLE IF NOT EXISTS
            {table} (
                subject TEXT NOT NULL,
                timestamp TIMESTAMP NOT NULL,
                {params}
            )
            """
        ).strip()
        await db.execute(stmt)

    async def insert_values(
        self, db: Connection, subject: Subject, model: ParametricModel
    ) -> None:
        # Here also, static prepared statements cannot be used to insert a dynamic
        # number of columns, so we dynamically create the prepared statement.
        table = self.safe_table_name(model.name)
        timestamp = datetime.now(timezone.utc)
        params_placeholder = ", ".join(["?"] * model.parameters.shape[-1])
        stmt = dedent(
            f"""
            INSERT INTO {table}
            VALUES (
                ?,
                ?,
                {params_placeholder}
            )
            """  # noqa: S608
        ).strip()
        args = [(subject, timestamp, *params) for params in model.parameters]
        await db.executemany(stmt, args)
        await db.commit()

    async def get_model_tables(self, db: Connection) -> list[str]:
        async with db.execute(
            "SELECT name FROM sqlite_master WHERE type='table'"
        ) as cursor:
            results = await cursor.fetchall()
            return [row[0] for row in results]

    async def select_latest_model(
        self, db: Connection, table: str
    ) -> dict[Subject, ParametricModel]:
        stmt = dedent(
            f"""
            SELECT
                *
            FROM
                {table}
            WHERE
                (subject, timestamp) IN (
                    SELECT
                        subject,
                        MAX(timestamp)
                    FROM
                        {table}
                    GROUP BY
                        subject
                )
            ORDER BY subject
            """  # noqa: S608
        ).strip()
        async with db.execute(stmt) as cursor:
            # TODO This could be more memory-efficient by returning an asynchronous
            # iterator instead.
            rows = await cursor.fetchall()
            return {
                subject: ParametricModel(
                    name=table,
                    # TODO Is the row order guaranteed to be the same as the insertion
                    # order ?
                    parameters=np.array([row[2:] for row in subject_rows]),
                )
                for subject, subject_rows in groupby(rows, key=lambda x: x[0])
            }



import logging
import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path

from typing_extensions import Self

from lisa.globalfit.framework.msg.base import MessageBase
from lisa.globalfit.framework.msg.subjects import Subject

logger = logging.getLogger(__name__)


@dataclass
class Response:
    subject: Subject
    msg: MessageBase


@dataclass
class Rule(ABC):
    id: str
    is_active: bool
    subject: str

    @abstractmethod
    def evaluate(self, subject: str, msg: bytes) -> list[Response]:
        pass

    @classmethod
    def from_file(cls, file: Path, variable_name: str = "__rule__") -> Self:
        # TODO Add random salt to imported module names to prevent collisions
        spec = spec_from_file_location(file.with_suffix("").name, file)
        if spec is None or spec.loader is None:
            raise ImportError(f"could not load rule from {file}")

        module = module_from_spec(spec)
        sys.modules[module.__name__] = module
        # This is potentially very dangerous, as it allows abitrary code execution when
        # loading rules.
        # Strict access control mechanisms need to be enforced for rule directories.
        spec.loader.exec_module(module)
        attr = getattr(module, variable_name, None)
        if attr is None:
            raise ImportError(f"could not find {variable_name!r} attribute in {file}")

        if not isinstance(attr, cls):
            raise TypeError(
                f"expected {cls.__name__} but got {attr.__class__.__name__}"
            )

        return attr



from operator import itemgetter
from pathlib import Path
from tempfile import NamedTemporaryFile

from lisa.globalfit.framework.monitoring.types import MonitoringStorage
from lisa.globalfit.framework.msg.subjects import (
    SUBJECT_GLOBALFIT,
    channel_get_group,
    channel_get_module,
    channel_get_pipeline,
)
from lisa.globalfit.framework.plot import ModuleExecutionDrawer


async def get_runs(storage: MonitoringStorage):
    runs = {}
    paginator = storage.client.get_paginator("list_objects_v2")
    async for result in paginator.paginate(
        Bucket=storage.bucket,
        Prefix="lisa-checkpoint",
    ):
        for item in result.get("Contents", []):
            run_id = channel_get_pipeline(item["Key"])
            last_modified = item["LastModified"]
            if run_id not in runs:
                runs[run_id] = {"run_id": run_id, "date": last_modified}
            else:
                runs[run_id]["date"] = max(runs[run_id]["date"], last_modified)

    return {"runs": sorted(runs.values(), key=itemgetter("date"), reverse=True)}


async def get_run_groups(storage: MonitoringStorage, run_id: str):
    groups = {}
    paginator = storage.client.get_paginator("list_objects_v2")
    async for result in paginator.paginate(
        Bucket=storage.bucket,
        Prefix=f"lisa-checkpoint-{SUBJECT_GLOBALFIT}.{run_id}",
    ):
        for item in result.get("Contents", []):
            group = channel_get_group(item["Key"])
            module = channel_get_module(item["Key"])
            if group not in groups:
                groups[group] = {"group_id": group, "modules": {module}}
            else:
                # mypy does not understand the in-place assignment,
                # so we split it in three steps
                group_modules = set(groups[group]["modules"])
                group_modules.add(module)
                groups[group]["modules"] = group_modules

    # Sort module names within each group.
    for group_id, item in groups.items():
        groups[group_id]["modules"] = sorted(item["modules"])

    return {"groups": sorted(groups.values(), key=itemgetter("group_id"))}


async def get_run_group_modules(storage: MonitoringStorage, run_id: str, group_id: str):
    paginator = storage.client.get_paginator("list_objects_v2")
    return sorted(
        {
            channel_get_module(item["Key"])
            async for result in paginator.paginate(
                Bucket=storage.bucket,
                Prefix=f"lisa-checkpoint-{SUBJECT_GLOBALFIT}.{run_id}.{group_id}",
            )
            for item in result.get("Contents", [])
        }
    )


async def get_run_group_module(
    storage: MonitoringStorage, run_id: str, group_id: str, module_id: str
):
    response = await storage.client.get_object(
        Bucket=storage.bucket,
        Key=f"lisa-checkpoint-{SUBJECT_GLOBALFIT}.{run_id}.{group_id}.{module_id}",
    )

    # TODO Refactor into self-contained functions.
    with (
        NamedTemporaryFile() as input_data,
        NamedTemporaryFile(suffix=".svg") as output_img,
    ):
        async for chunk in response["Body"].iter_chunks():
            input_data.write(chunk)
        input_data.flush()

        ModuleExecutionDrawer().plot_corner(
            Path(input_data.name), "mcmc/chain", Path(output_img.name)
        )
        return {"plots": [{"src": output_img.read().decode()}]}



from dataclasses import dataclass
from typing import Optional

from typing_extensions import Self

from lisa.globalfit.framework.model import MarkovChainSamples, ParametricModel
from lisa.globalfit.framework.msg.base import MessageBase


@dataclass
class DataMessage(MessageBase):
    """Message for passing metadata between modules."""


@dataclass
class ModuleState(DataMessage):
    """Comprehensive state of a module.

    This state can be used as module checkpoint, from which MCMC sampling can be
    resumed.

    :param chain: Samples leading to current state.
    """

    chain: MarkovChainSamples
    log_prior: ParametricModel
    log_likelihood: ParametricModel
    log_proposal: ParametricModel
    checkpoint_key: Optional[str] = None

    @classmethod
    def from_dict(cls, obj: dict) -> Self:
        return cls(
            chain=MarkovChainSamples.from_dict(obj["chain"]),
            log_prior=ParametricModel.from_dict(obj["log_prior"]),
            log_likelihood=ParametricModel.from_dict(obj["log_likelihood"]),
            log_proposal=ParametricModel.from_dict(obj["log_proposal"]),
            checkpoint_key=obj["checkpoint_key"],
        )


@dataclass
class ErrorMessage(DataMessage):
    """Message for reporting errors."""

    pass



from argparse import ArgumentParser
from argparse import _SubParsersAction as Subparser
from pathlib import Path

from lisa.globalfit.framework.logging import DEFAULT_LOG_LEVEL, LOG_LEVELS


def add_common_args(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--log-level",
        choices=LOG_LEVELS,
        default=DEFAULT_LOG_LEVEL,
        help="set the logging level",
    )
    parser.add_argument(
        "--workdir",
        help="Working directory",
        type=Path,
    )


def add_bus_args(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--servers",
        nargs="+",
        default=["nats://localhost:4222"],
        help="event bus server URLs",
    )


def add_pipeline_args(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--pipeline-run-id",
        help="pipeline run identifier",
        type=str,
        required=True,
    )


def add_preprocessing_args(parser: Subparser) -> None:
    subparser: ArgumentParser = parser.add_parser(
        "preprocess",
        help="run preprocessors on input dataset",
    )
    subparser.add_argument(
        "--group-label",
        type=str,
        help="label for modules of this group",
        required=True,
    )
    subparser.add_argument(
        "--input-data",
        type=Path,
        help="input dataset",
        required=True,
    )
    subparser.add_argument(
        "--input-checkpoint",
        type=Path,
        help="checkpoint of previous iteration",
    )
    subparser.add_argument(
        "--input-configuration-file",
        "-c",
        type=Path,
        help="path to configuration file",
    )
    subparser.add_argument(
        "--output-preprocessed-signal",
        type=Path,
        help="pre-processed input signal forwarded to modules",
        required=True,
    )
    subparser.add_argument(
        "--output-group-configuration-path",
        type=Path,
        default=Path("config_group.json"),
        help="configuration file of group",
    )


def add_modules_execution_args(parser: Subparser) -> None:
    subparser: ArgumentParser = parser.add_parser(
        "start-modules",
        help="spawn processing modules",
    )
    subparser.add_argument(
        "--input-group-configuration-path",
        type=Path,
        help="group configuration file",
        required=True,
    )
    subparser.add_argument(
        "--input-data",
        type=Path,
        help="input dataset",
        required=True,
    )
    subparser.add_argument(
        "--current-iteration",
        help="current iteration index",
        type=int,
        required=True,
    )
    subparser.add_argument(
        "--n-steps",
        type=int,
        help="number of MCMC steps to perform",
        default=1000,
    )


def add_module_execution_args(parser: Subparser) -> None:
    subparser: ArgumentParser = parser.add_parser(
        "start-module",
        help="spawn a single processing module",
    )
    subparser.add_argument(
        "--input-data",
        type=Path,
        help="input dataset",
        required=True,
    )
    subparser.add_argument(
        "--group-label",
        help="group label",
        type=str,
        required=True,
    )
    subparser.add_argument(
        "--module-label",
        help="module label",
        type=str,
        required=True,
    )
    subparser.add_argument(
        "--input-group-configuration-path",
        help="module initial configuration",
        type=Path,
        required=True,
    )
    subparser.add_argument(
        "--input-configuration-file",
        "-c",
        type=Path,
        help="path to configuration file",
    )
    subparser.add_argument(
        "--current-iteration",
        help="current iteration index",
        type=int,
        required=True,
    )
    subparser.add_argument(
        "--n-steps",
        type=int,
        help="number of MCMC steps to perform",
        default=1000,
    )
    subparser.add_argument(
        "--output-checkpoint",
        type=Path,
        help="module state needed to resume execution",
        required=True,
    )


def add_engine_execution_args(parser: Subparser) -> None:
    subparser: ArgumentParser = parser.add_parser(
        "start-engine",
        help="start framework engine",
    )
    subparser.add_argument(
        "--expected-groups",
        type=Path,
        help="expected worker groups",
        required=True,
    )
    subparser.add_argument(
        "--output-checkpoint",
        type=Path,
        help="output iteration checkpoint",
        default=Path("checkpoint.json"),
    )
    subparser.add_argument(
        "--output-iteration-decision",
        type=Path,
        help="iteration decision from rules evaluation",
        required=True,
    )
    subparser.add_argument(
        "--catalog-db",
        type=Path,
        help="local catalog of detected sources",
    )
    subparser.add_argument(
        "--max-iterations",
        type=int,
        help="maximum number of iterations",
        required=True,
    )
    subparser.add_argument(
        "--current-iteration",
        type=int,
        help="current iteration number",
        required=True,
    )


def add_execution_plot_args(parser: Subparser) -> None:
    subparser: ArgumentParser = parser.add_parser(
        "plot",
        help="plot MCMC results",
    )
    subparser.add_argument(
        "--input-chain",
        type=Path,
        help="path to an HDF5 file containing sampler chains",
        required=True,
    )
    subparser.add_argument(
        "--chain-location",
        type=str,
        default="mcmc/chain",
        help="path to chain dataset in input file",
    )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# Project information
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "GlobalFit Framework"
copyright = "2024, CNES"
author = "CNES"

# General configuration
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ["_templates"]
exclude_patterns = []


# Options for HTML output
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]



import logging
from typing import Callable, Optional

from nats.aio.client import Client
from nats.aio.msg import Msg
from nats.aio.subscription import Subscription
from nats.js import JetStreamContext

from lisa.globalfit.framework.bus import EventBus
from lisa.globalfit.framework.exceptions import FrameworkError
from lisa.globalfit.framework.msg.subjects import PERSISTENT_SUBJECTS, Subject

logger = logging.getLogger(__name__)


class NatsEventBus(EventBus):
    DEFAULT_STREAM = next(iter(__name__.split(".", maxsplit=1)))
    DEFAULT_SUBJECTS = PERSISTENT_SUBJECTS

    def __init__(self, servers: list[str]) -> None:
        super().__init__()
        self.servers = servers
        self.client = Client()
        self.stream: Optional[JetStreamContext] = None
        self.subscription_handles: dict[Subject, Subscription] = {}

    async def connect(
        self, stream: str = DEFAULT_STREAM, subjects: list[Subject] = DEFAULT_SUBJECTS
    ):
        logger.debug(f"connecting to NATS: {self.servers}")
        await self.client.connect(servers=self.servers)

        logger.debug("creating jetstream context")
        # FIXME Setting a None timeout is required to perform lengthy operations (such
        # as MCMC iterations) in asynchronous subscription callbacks without being
        # timed-out by the NATS client. There might be a better way to do that.
        self.stream = self.client.jetstream(timeout=None)

        logger.debug(f"adding stream '{stream}' with subjects {subjects}")
        subject_wildcards = [s + ".>" for s in subjects]
        await self.stream.add_stream(name=stream, subjects=subject_wildcards)
        logger.debug(f"connected to NATS: {await self.stream.streams_info()}")

    async def publish(self, subject: Subject, data: bytes) -> None:
        if self.stream is None:
            raise FrameworkError("not connected to NATS")
        await self.stream.publish(subject, data)
        logger.debug(f"published {len(data)} bytes on subject {subject}")

    async def subscribe(self, subject: Subject, callback: Callable) -> None:
        if self.stream is None:
            raise FrameworkError("not connected to NATS")

        if subject in self.subscription_handles:
            msg = f"multiple subscriptions to {subject} are not supported"
            raise FrameworkError(msg)

        async def message_handler(msg: Msg):
            await callback(msg.subject, msg.data)

        subscriber = await self.stream.subscribe(subject, cb=message_handler)
        self.subscription_handles[subject] = subscriber
        logger.debug(f"got subscription from {callback.__qualname__} to '{subject}'")

    async def unsubscribe(self, subject: Subject) -> None:
        try:
            subscription = self.subscription_handles[subject]
        except KeyError as err:
            logger.warning(f"no subscription for {err}")
            return

        await subscription.unsubscribe()

    async def close(self) -> None:
        await self.client.drain()



# Import all modules so that they can register themselvses to the registry.
from lisa.globalfit.framework.modules.stub import module

# Remove the imported modules once their code have run to prevent unused import
# warnings.
del module



import os
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass

from aiobotocore.session import get_session
from types_aiobotocore_s3 import S3Client
from typing_extensions import Self


@dataclass
class MonitoringStorage:
    client: S3Client
    bucket: str

    @classmethod
    @asynccontextmanager
    async def from_env(cls) -> AsyncIterator[Self]:
        async with get_session().create_client("s3") as s3:
            yield cls(s3, os.environ["S3_BUCKET"])



import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from pathlib import Path

from lisa.globalfit.framework.simulator import Simulator


def add_simulator_args(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--n-gaussians",
        type=int,
        default=2,
        help="number of gaussians to sum",
    )
    parser.add_argument(
        "--n-samples",
        type=int,
        default=365,
        help="number of data points to simulate",
    )
    parser.add_argument(
        "--snr",
        type=int,
        default=20,
        help="signal to noise ratio",
    )
    parser.add_argument(
        "--output-path",
        type=Path,
        default=Path("simulated_signal.h5"),
        help="output file to write the simulated data to",
    )


def get_args() -> Namespace:
    parser = ArgumentParser(
        "Simulate LISA-like signals",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    add_simulator_args(parser)
    return parser.parse_args()


def cmd_simulate_signal(args: Namespace):
    sim = Simulator()
    signal = sim.simulate_gaussians(args.n_gaussians, args.n_samples, args.snr)
    signal.write_dataset(args.output_path)
    signal.plot(args.output_path.with_suffix(".png"))


def main():
    args = get_args()
    cmd_simulate_signal(args)


if __name__ == "__main__":
    sys.exit(main())



import asyncio
import json
import logging
import subprocess
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field

from lisa.globalfit.framework.args import add_bus_args, add_common_args
from lisa.globalfit.framework.buses.nats import NatsEventBus
from lisa.globalfit.framework.engine.rule import Response, Rule
from lisa.globalfit.framework.msg.base import MessageBase

logger = logging.getLogger(__name__)

INBOX = "infra.pipelinerunner.submit"
OUTBOX = "infra.pipelinerunner.result"


@dataclass
class WorkflowSpec(MessageBase):
    resource: str
    labels: list[str] = field(default_factory=list)


@dataclass
class WorkflowResult(MessageBase):
    status: str
    labels: list[str] = field(default_factory=list)


class SubmitWorkflow(Rule):
    def evaluate(self, subject: str, msg: bytes) -> list[Response]:
        if subject != INBOX:
            logger.warning(f"got workflow submission from unexpected subject {subject}")
            return []

        spec = WorkflowSpec.decode(msg)
        result = self.submit_workflow(spec)

        return [Response(OUTBOX, result)]

    def submit_workflow(self, spec: WorkflowSpec) -> WorkflowResult:
        logger.info(f"submitting workflow {spec}")
        proc = subprocess.run(
            ["argo", "submit", "--output", "name", "--wait", spec.resource],
            capture_output=True,
        )
        if proc.returncode != 0:
            raise ValueError(
                f"could not submit workflow: {proc.stderr.decode().strip()}"
            )

        workflow_name = proc.stdout.decode().strip()

        return self.get_workflow_status(spec, workflow_name)

    def get_workflow_status(self, spec: WorkflowSpec, name: str) -> WorkflowResult:
        logger.info(f"fetching status of workflow {name!r}")
        proc = subprocess.run(
            ["argo", "get", "--output", "json", name],
            capture_output=True,
        )
        if proc.returncode != 0:
            raise ValueError(
                f"could not get workflow status: {proc.stderr.decode().strip()}"
            )

        contents = json.loads(proc.stdout)
        return WorkflowResult(contents["status"]["phase"], spec.labels)


__rule__ = SubmitWorkflow(id="pipelinerunner-bridge", is_active=True, subject=INBOX)


def get_args() -> Namespace:
    parser = ArgumentParser()

    add_common_args(parser)
    add_bus_args(parser)

    parser.add_argument(
        "workflow",
        type=str,
    )

    parser.add_argument(
        "--labels",
        type=str,
        nargs="+",
    )

    return parser.parse_args()


async def main_async():
    args = get_args()
    logging.basicConfig(
        level=args.log_level,
        format="%(asctime)s.%(msecs)03d [%(levelname)s] %(name)s %(message)s",
        datefmt=r"%Y-%m-%dT%H:%M:%S",
    )
    bus = NatsEventBus(servers=args.servers)
    spec = WorkflowSpec(
        resource=args.workflow, labels=args.labels if args.labels is not None else []
    )
    await bus.connect()
    await bus.publish(INBOX, spec.encode())
    logger.info(f"submitted workflow {spec}")


def main():
    return asyncio.run(main_async())


if __name__ == "__main__":
    main()



import logging
from abc import abstractmethod
from dataclasses import dataclass
from typing import ClassVar

import numpy as np

logger = logging.getLogger(__name__)


@dataclass
class Waveform:
    """Abstract base class representing a waveform."""

    @abstractmethod
    def evaluate(self, x: np.ndarray, *args: float) -> np.ndarray:
        """Evaluate the waveform at given points.

        :param x: Array of points at which to evaluate the waveform.
        :return: Array containing the values of the waveform at the given points.
        """


@dataclass
class WaveformRegistry:
    items: ClassVar[dict[str, type[Waveform]]] = {}

    @classmethod
    def choices(cls) -> list[str]:
        return list(cls.items)

    @classmethod
    def register(cls, name: str):
        def decorator(waveform_class: type[Waveform]):
            cls.items[name] = waveform_class
            return waveform_class

        return decorator

    @classmethod
    def create(cls, name: str, *args, **kwargs) -> Waveform:
        try:
            kls = cls.items[name]
            return kls(*args, **kwargs)
        except KeyError:
            logger.error(f"unknown waveform identifier: {name}")
            raise


@dataclass
@WaveformRegistry.register("gaussian1d")
class Gaussian1D(Waveform):
    def evaluate(self, x: np.ndarray, *args: float) -> np.ndarray:
        mean, variance = args
        return np.exp(-0.5 * ((x - mean) / variance) ** 2)



from aiohttp import ClientSession
from aiohttp.web import (
    AppKey,
    Application,
    Request,
    Response,
    RouteTableDef,
)
from jinja2 import Environment, PackageLoader, select_autoescape

jinja_env = AppKey("jinja_env", Environment)
api_session = AppKey("api_session", ClientSession)


def render_template(request: Request, filename: str, **kwargs) -> Response:
    template = request.app[jinja_env].get_template(filename)
    return Response(
        text=template.render(**kwargs),
        content_type="text/html",
    )


async def api_get_json(request: Request, path: str) -> dict:
    res = await request.app[api_session].get(f"/api{path}")
    return await res.json()


routes = RouteTableDef()


@routes.get("/")
async def get_root(request: Request) -> Response:
    return render_template(
        request,
        "index.html",
        navigation=[{"caption": "Pipeline runs", "href": "runs"}],
    )


@routes.get("/runs/")
async def get_runs(request: Request) -> Response:
    return render_template(
        request,
        "runs.html",
        **await api_get_json(request, "/runs"),
    )


@routes.get("/runs/{run}/")
async def get_run_details(request: Request) -> Response:
    groups = await api_get_json(request, f"/runs/{request.match_info['run']}/groups")
    return render_template(
        request,
        "run_details.html",
        **groups,
    )


@routes.get("/runs/{run}/groups/{group}/modules/{module}")
async def get_module_details(request: Request) -> Response:
    details = await api_get_json(
        request,
        "/".join(
            (
                "/runs",
                request.match_info["run"],
                "groups",
                request.match_info["group"],
                "modules",
                request.match_info["module"],
            )
        ),
    )
    return render_template(
        request,
        "module_details.html",
        **details,
    )


def register(app: Application, prefix: str, api_url: str):
    subapp = Application()
    subapp.add_routes(routes)

    # Configure Jinja.
    subapp[jinja_env] = Environment(
        loader=PackageLoader(__name__, "templates"),
        autoescape=select_autoescape(),
    )
    subapp[api_session] = ClientSession(base_url=api_url)

    app.add_subapp(prefix, subapp)



import logging
from functools import partial
from typing import Callable, Optional

import numpy as np
from emcee import EnsembleSampler, State
from emcee.backends import HDFBackend
from scipy import stats

from lisa.globalfit.framework.exceptions import FrameworkError
from lisa.globalfit.framework.model import (
    MarkovChainSamples,
    ModuleExecutionState,
    ParametricModel,
)
from lisa.globalfit.framework.modules.module import ModuleBase
from lisa.globalfit.framework.modules.registry import ModuleRegistry
from lisa.globalfit.framework.msg.control import ConfigureModule, IterateModule
from lisa.globalfit.framework.msg.data import ModuleState
from lisa.globalfit.framework.signal import NoisyGaussianMixture
from lisa.globalfit.framework.waveform import Waveform, WaveformRegistry

logger = logging.getLogger(__name__)


def log_likelihood(wave: Waveform, sample: np.ndarray, signal: np.ndarray) -> float:
    waveform = wave.evaluate(signal[0], *sample)
    residuals = signal[1] - waveform
    square_err = np.sum(residuals**2)
    return -np.log(square_err)


def log_prob(
    prior_fn: Callable,
    log_likelihood_fn: Callable,
    sample: np.ndarray,
    dataset: NoisyGaussianMixture,
) -> float:
    log_prior = prior_fn(sample)
    log_likelihood = log_likelihood_fn(sample, dataset)
    return log_prior + log_likelihood


@ModuleRegistry.register("stub")
class StubModule(ModuleBase):
    """Placeholder module."""

    DEFAULT_SAMPLER_WALKERS = 5

    def __init__(
        self,
        nwalkers: int = DEFAULT_SAMPLER_WALKERS,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.dataset = NoisyGaussianMixture.read_dataset(self.datafile)
        self.n_walkers = nwalkers
        self.rng = np.random.default_rng()
        self.sampler: Optional[EnsembleSampler] = None
        self.sampler_state: Optional[State] = None
        self.log_prob: Optional[Callable] = None
        self.waveform: Optional[Waveform] = None
        self.configure_message: Optional[ConfigureModule] = None

    async def configure(self, msg: ConfigureModule) -> None:
        # Save configure message to build iteration message.
        self.configure_message = msg

        # Set-up prior.
        prior_key = msg.state.log_prior.name
        try:
            dist = getattr(stats, msg.state.log_prior.name)
            logpdf = getattr(dist, "logpdf")
        except AttributeError:
            logger.error(f"prior '{prior_key}' not in known distributions")
            await self.update_status(ModuleExecutionState.ERROR)
            return

        # Build posterior evaluation function using Bayes theorem.
        self.waveform = WaveformRegistry.create(msg.state.log_likelihood.name)
        log_prior_fn = partial(logpdf, msg.state.log_prior.parameters)
        log_likelihood_fn = partial(log_likelihood, self.waveform)
        self.log_prob = partial(log_prob, log_prior_fn, log_likelihood_fn)

        initial_guess = np.repeat(
            msg.state.chain.waveforms.parameters, self.n_walkers, axis=0
        )
        # TODO Initial state need to be scattered for walkers to be independent. Is
        # this the responsibility of the pre-processing or should we do it here ?
        initial_guess += self.rng.normal(size=initial_guess.shape)
        self.sampler_state = State(coords=initial_guess)

        self.sampler = EnsembleSampler(
            nwalkers=self.n_walkers,
            ndim=initial_guess.shape[-1],
            log_prob_fn=self.log_prob,
        )

    async def iterate(self, control: IterateModule) -> ModuleState:
        if (
            self.dataset is None
            or self.sampler is None
            or self.sampler_state is None
            or self.waveform is None
        ):
            raise FrameworkError("trying to iterate misconfigured module")

        # Subtract other sources from signal.
        module_signal = self.dataset.signal.copy()
        params = await control.detections.get_source_parameters()
        for module_name, model in params.items():
            if module_name == self.channel_name:
                continue

            # TODO Handle waveforms of other module types.
            module_signal[1] -= self.waveform.evaluate(
                module_signal[0], *model.parameters[0]
            )

        # This is awfully slow on shared filesystems, but is only used for integration
        # purposes.
        # See https://github.com/dfm/emcee/issues/393
        backend = HDFBackend(self.output_checkpoint)

        # Re-create the sampler, because we cannot just change the posterior args
        # on-the-fly.
        self.sampler = EnsembleSampler(
            nwalkers=self.sampler.nwalkers,
            ndim=self.sampler.ndim,
            log_prob_fn=self.log_prob,
            args=[module_signal],
            backend=backend,
        )

        # Run MCMC on residual.
        new_state = self.sampler.run_mcmc(
            initial_state=self.sampler_state,
            nsteps=control.step_count,
        )
        if new_state is None:
            msg = f"{self.channel_name} got invalid state {new_state} "
            f"at iteration {self.current_iteration}"
            raise FrameworkError(msg)

        self.sampler_state = new_state
        return self.build_iteration_message()

    def build_iteration_message(self) -> ModuleState:
        if self.sampler is None or self.configure_message is None:
            raise FrameworkError("trying to build step message without sampler state")

        flatlnprob = self.sampler.get_log_prob(flat=True)
        flatchain = self.sampler.get_chain(flat=True)
        if flatlnprob is None or flatchain is None:
            raise FrameworkError("trying to build step message without posterior")

        best_sample_pos = np.argmax(flatlnprob)
        best_sample_prob = flatlnprob[best_sample_pos]
        best_sample = flatchain[best_sample_pos]
        logger.info(
            f"{self.channel_name} iteration {self.current_iteration} best sample: "
            f"{best_sample} logprob: {best_sample_prob}"
        )

        return ModuleState(
            chain=MarkovChainSamples(
                idx_iteration=self.current_iteration,
                idx_step_end=self.current_step,
                waveforms=ParametricModel(
                    name=self.configure_message.state.log_likelihood.name,
                    parameters=np.array([best_sample]),
                ),
            ),
            log_prior=self.configure_message.state.log_prior,
            log_proposal=self.configure_message.state.log_proposal,
            log_likelihood=self.configure_message.state.log_likelihood,
        )



from abc import ABC, abstractmethod
from pathlib import Path

from lisa.globalfit.framework.model import ParametricModel
from lisa.globalfit.framework.msg.data import ModuleState
from lisa.globalfit.framework.msg.subjects import Subject


class SourceCatalog(ABC):
    """Gravitational wave detection catalog."""

    def __init__(self) -> None:
        pass

    @abstractmethod
    async def update(self, module: Subject, state: ModuleState) -> None:
        pass

    @abstractmethod
    async def get_source_parameters(self) -> dict[Subject, ParametricModel]:
        pass

    @abstractmethod
    async def write(self, filename: Path) -> None:
        pass



import json
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from lisa.globalfit.framework.catalog import SourceCatalog
from lisa.globalfit.framework.io import CustomJSONEncoder
from lisa.globalfit.framework.model import ParametricModel
from lisa.globalfit.framework.msg.data import ModuleState
from lisa.globalfit.framework.msg.subjects import Subject


@dataclass
class InMemorySourceCatalog(SourceCatalog):
    def __init__(
        self, detection_map: Optional[dict[Subject, ParametricModel]] = None
    ) -> None:
        super().__init__()
        self.detection_map = detection_map or {}

    async def get_source_parameters(self) -> dict[Subject, ParametricModel]:
        return self.detection_map

    async def update(self, module: Subject, state: ModuleState) -> None:
        # TODO Should the detection catalog be updated if the best sample from the
        # chain in the current iteration has a lower posterior probability than the
        # best sample from all previous iterations ?
        # The posterior should probably be passed along the waveform parameters, so
        # that its evolution can be monitored during execution.
        self.detection_map[module] = state.chain.waveforms

    async def write(self, filename: Path) -> None:
        filename.write_text(json.dumps(self, cls=CustomJSONEncoder))



import json
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Callable

from typing_extensions import Self

from lisa.globalfit.framework.io import CustomJSONEncoder
from lisa.globalfit.framework.model import ModuleExecutionState
from lisa.globalfit.framework.msg.control import ExecutionStateUpdate
from lisa.globalfit.framework.msg.data import ModuleState
from lisa.globalfit.framework.msg.subjects import Subject, create_channel_name

logger = logging.getLogger(__name__)


@dataclass
class ModulePoolEntry:
    group: str
    label: str
    execution_state: ModuleExecutionState
    state: ModuleState
    current_iteration: int = -1

    @classmethod
    def from_dict(cls, obj: dict) -> Self:
        return cls(
            group=obj["group"],
            label=obj["label"],
            execution_state=ModuleExecutionState(obj["execution_state"]),
            state=ModuleState.from_dict(obj["state"]),
            current_iteration=obj["current_iteration"],
        )


@dataclass
class ModulePool:
    workers: dict[Subject, ModulePoolEntry]

    def select_modules(
        self,
        checker: Callable[[ModulePoolEntry], bool],
    ) -> list[Subject]:
        return [subject for subject, entry in self.workers.items() if checker(entry)]

    async def update_module_state(
        self,
        module: Subject,
        update: ExecutionStateUpdate,
    ) -> None:
        # TODO Add sanity checks and log messages.
        if module not in self.workers:
            logger.error(f"trying to update status of module {module} not in pool")
            return

        entry = self.workers[module]
        logger.debug(f"module {module}: {entry.execution_state} -> {update.new_state}")

        entry.execution_state = update.new_state
        entry.current_iteration = update.current_iteration

    def write_checkpoint(self, file: Path) -> None:
        with open(file, "w") as f:
            json.dump(self, f, cls=CustomJSONEncoder)
        logger.info(f"wrote module pool checkpoint {file}")

    def __len__(self):
        return len(self.workers)

    @classmethod
    def from_json(cls, file: Path) -> Self:
        return cls(
            workers={
                subject: ModulePoolEntry.from_dict(module)
                for subject, module in json.loads(file.read_text())["workers"].items()
            }
        )

    @classmethod
    def from_group_configuration(cls, file: Path, *channel_args: str) -> Self:
        return cls(
            workers={
                create_channel_name(
                    *channel_args,
                    module["group_label"],
                    module["module_label"],
                ): ModulePoolEntry(
                    label=module["module_label"],
                    group=module["group_label"],
                    execution_state=ModuleExecutionState.CREATED,
                    state=ModuleState.from_dict(module["initial_state"]),
                )
                for module in json.loads(file.read_text())
            },
        )

    async def update_global_state(self, module: Subject, state: ModuleState) -> None:
        if module not in self.workers:
            logger.error(f"trying to update state of module {module} not in pool")
            return

        if not state.checkpoint_key:
            logger.warning(f"missing checkpoint key for module {module}")

        logger.debug(f"updating global state with results from {module}")
        self.workers[module].state = state



from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path

import h5py
import numpy as np
from h5py import Dataset
from matplotlib import pyplot as plt
from typing_extensions import Self


@dataclass
class Signal:
    @abstractmethod
    def write_dataset(self, output: Path) -> None:
        pass

    @abstractmethod
    def plot(self, file: Path) -> None:
        pass

    @classmethod
    @abstractmethod
    def read_dataset(cls, file: Path) -> Self:
        pass


def read_dataset_array(file: h5py.File, path: str) -> np.ndarray:
    """Read a dataset from an HDF5 file to an in-memory array.

    :param file: File to read from.
    :param path: Path to dataset in file.
    :raises ValueError: If the path does not exist in the file or is not a dataset.
    :return: Array containing dataset values.
    """
    dataset = file.get(path)
    if not isinstance(dataset, Dataset):
        raise ValueError(f"invalid {path} dataset ({dataset})")
    return np.array(dataset[:])


@dataclass
class NoisyGaussianMixture(Signal):
    """Sum of gaussians with noise.

    :param means: Array of shape (n_distributions,) with true mean values.
    :param variances: Array of shape (n_distributions,) with true variance values.
    :param noise: Array of shape (n_samples,) with noise values. Implicitly uses the
        same time points as the signal.
    :param signal: Array of shape (2, n_samples) with time points and noisy signal
        values.
    """

    means: np.ndarray
    variances: np.ndarray
    noise: np.ndarray
    signal: np.ndarray

    def n_distributions(self) -> int:
        """Count the number of distributions in the signal.

        :return: Number of distributions.
        """
        return len(self.means)

    def write_dataset(self, output: Path) -> None:
        """Write the signal to a HDF5 file.

        :param output: File to write to.
        """
        with h5py.File(output, "w") as f:
            catalog = f.create_group("cat")
            catalog.create_dataset("means", data=self.means)
            catalog.create_dataset("variances", data=self.variances)

            instrument = f.create_group("instrument")
            instrument.create_dataset("noise", data=self.noise)

            observations = f.create_group("obs")
            observations.create_dataset("signal", data=self.signal)

    def plot(self, file: Path) -> None:
        """Plot the signal to an image file.

        :param file: File to write image to.
        """
        plt.figure()
        time_points, signal = self.signal
        plt.plot(time_points, signal, label="noisy signal", color="darkgrey")
        plt.plot(time_points, signal - self.noise, label="ideal data", color="tab:blue")
        plt.vlines(
            self.means,
            ymin=0,
            ymax=signal.max(),
            label="means",
            linestyles="--",
            color="tab:red",
        )
        plt.legend(loc="best")
        plt.savefig(file)
        plt.close()

    @classmethod
    def read_dataset(cls, file: Path) -> Self:
        """Read a dataset from a HDF5 file.

        :param file: File to read from.
        """
        with h5py.File(file, "r") as f:
            return cls(
                means=read_dataset_array(f, "cat/means"),
                variances=read_dataset_array(f, "cat/variances"),
                noise=read_dataset_array(f, "instrument/noise"),
                signal=read_dataset_array(f, "obs/signal"),
            )



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from lisa.globalfit.framework.catalog import SourceCatalog
from lisa.globalfit.framework.model import ModuleExecutionState
from lisa.globalfit.framework.msg.base import MessageBase
from lisa.globalfit.framework.msg.data import ModuleState


@dataclass
class ControlMessage(MessageBase):
    """Message for controlling Global Fit flow."""

    pass


@dataclass
class ConfigureModule(ControlMessage):
    """Configure an initialized module."""

    state: ModuleState
    configuration_file: Optional[Path] = None


@dataclass
class IterateModule(ControlMessage):
    """Iterate a module for a given number of steps.

    :param idx_iteration: Current iteration counter.
    :param detections: Live catalog.
    :param step_count: Number of steps to execute in iteration.
    """

    idx_iteration: int
    detections: SourceCatalog
    step_count: int


@dataclass
class ContinueModule(ControlMessage):
    """Finish the current step and wait for next iteration order."""

    pass


@dataclass
class TerminateModule(ControlMessage):
    """Finish the current step and exit."""

    pass


@dataclass
class KillModule(ControlMessage):
    """Exit immediately."""

    pass


@dataclass
class ExecutionStateUpdate(ControlMessage):
    """Transition of execution state."""

    current_iteration: int
    new_state: ModuleExecutionState


@dataclass
class ModuleGroupConfiguration(ControlMessage):
    """Configuration of a module group."""

    group: str
    labels: tuple[str, ...]



from pathlib import Path

import h5py
from corner import corner
from matplotlib import pyplot as plt

from lisa.globalfit.framework.signal import read_dataset_array


class ModuleExecutionDrawer:
    def plot_corner(self, chains: Path, dataset: str, output: Path) -> None:
        with h5py.File(chains) as f:
            samples = read_dataset_array(f, dataset)

        # Stack all walkers chains.
        samples = samples.reshape(
            samples.shape[0] * samples.shape[1], samples.shape[-1]
        )

        # Write corner plot to image.
        fig = corner(samples, show_titles=True, quantiles=[0.16, 0.5, 0.84])
        plt.savefig(output)
        plt.close(fig)



from dataclasses import dataclass
from pickle import PicklingError, UnpicklingError, dumps, loads

from typing_extensions import Self

from lisa.globalfit.framework.exceptions import DeserializationError, SerializationError


@dataclass
class MessageBase:
    """Base class for event messages exchanged within the framework."""

    def encode(self) -> bytes:
        """Encode the message for communication on the message bus.

        Implementations should return a binary representation of the message, using
        Python pickle serialization protocol.

        :return: Encoded message.
        """
        try:
            return dumps(self)
        except PicklingError as err:
            msg = f"could not encode message: {err}"
            raise SerializationError(msg) from err

    @classmethod
    def decode(cls, data: bytes) -> Self:
        """Decode a message from its binary representation.

        :param data: Binary content of the encoded message.

        :return: Decoded message.
        """
        # TODO: We would like to use a proper serialization format eventually.
        # However, protobuf does not support Numpy arrays natively, so it might
        # need a bit of work.
        try:
            return loads(data)  # noqa: S301
        except UnpicklingError as err:
            msg = f"could not decode message: {err}"
            raise DeserializationError(msg) from err



import json
import logging
from dataclasses import dataclass
from pathlib import Path

from lisa.globalfit.framework.io import CustomJSONEncoder
from lisa.globalfit.framework.msg.data import ModuleState

logger = logging.getLogger(__name__)


@dataclass
class ModuleConfiguration:
    group_label: str
    module_label: str
    initial_state: ModuleState
    static_configuration_reference: str


@dataclass
class PreprocessorOutput:
    modules: list[ModuleConfiguration]

    def write_json(self, output_file: Path) -> None:
        with open(output_file, "w", encoding="utf-8") as f:
            json.dump(self.modules, f, cls=CustomJSONEncoder)
        logger.info(f"wrote group configuration to {output_file}")



import sys
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, Namespace
from asyncio import run

from lisa.globalfit.framework.monitoring.server import MonitoringServer


def add_execution_monitoring_args(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--port",
        type=int,
        help="server port",
        default=8080,
    )
    parser.add_argument(
        "--api-url",
        type=str,
        help="URL of the backend service",
        default="http://localhost:8080",
    )
    parser.add_argument(
        "--s3-bucket",
        type=str,
        help="name of S3 bucket storing module checkpoints",
        required=True,
    )


def get_args() -> Namespace:
    parser = ArgumentParser(
        "LISA GlobalFit Framework monitoring CLI",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    add_execution_monitoring_args(parser)
    return parser.parse_args()


async def cmd_monitor(args: Namespace):
    server = MonitoringServer(args.api_url, args.s3_bucket)
    await server.run(args.port)


def main():
    args = get_args()
    run(cmd_monitor(args))


if __name__ == "__main__":
    sys.exit(main())
